%   Package: knuth_b_1.pl
%   Typist : Richard A. O'Keefe
%   Author : the rules of arithmetic, via Donald Knuth
%   Updated: 2/10/87
%   Purpose: table of constants.

/*  This is table 1 of appendix B of "The Art of Computer Programming"
    volume 1, by Donald E. Knuth.  That table gives 40 decimal places.
    As most of the Prolog systems known to me have trouble with 6
    places, I have rounded to 7 decimal places.  Anyone wanting more
    should consult the original reference.  Note that most of these
    names are compound terms, e.g. sqrt(2), 3^(1/3), cos(1).  This
    was done for clarity.

    This table may be extended with other constants taken from
    Abramowitz & Stegun, "Handbook of Mathematical Functions",
    chapter 1.  Note that they give numbers to 21 or 26 places.
    Which constants would be most useful to you?
*/

:- module(knuth_b_1, [
	constant/2
   ]).

constant(sqrt(2),	1.414124).
constant(sqrt(3),	1.732051).
constant(sqrt(5),	2.236068).
constant(sqrt(10),	3.162278).

constant(2^(1/3),	1.259921).
constant(3^(1/3),	1.442250).
constant(2^(1/4),	1.189207).

constant(ln(2),		0.6931472).
constant(ln(3),		1.098612).
constant(ln(10),	2.302585).
constant(1/ln(2),	1.442695).
constant(1/ln(10),	0.4342945).
constant(ln(ln(2)),    -0.3665129).

constant(pi,		3.141593).
constant(pi/180,	0.001745329).
constant(1/pi,		0.3183099).
constant(pi^2,		9.869604).
constant(sqrt(pi),	1.772454).
constant(ln(pi),	1.144730).

constant(gamma(1/2),	1.772454).
constant(gamma(1/3),	2.678938).
constant(gamma(2/3),	1.354118).

constant(e,		2.718282).
constant(1/e,		0.3678794).
constant(e^2,		7.389056).
constant(e^(pi/4),	2.193280).

constant(gamma,		0.5772157).
constant(e^gamma,	1.781072).

constant(phi,		1.618034).
constant(ln(phi),	0.4812118).
constant(1/ln(phi),	2.078087).

constant(sin(1),	0.8414710).
constant(cos(1),	0.5403023).

constant(zeta(3),	1.202057).

